/*
 * IGS vga chip register tool
 *
 */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <asm/segment.h>
//#include <unistd.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>

/*****************************************************************************/
#include "igsvga.h"

#define	SEQ_I	0x3C4
#define	SEQ_D	0x3C5

#define	CRT_I	0x3D4
#define	CRT_D	0x3D5


static int open_igsvga(struct inode* inodep, struct file* filep);
static void release_igsvga(struct inode* inodep, struct file* filep);
static int igsvga_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg);
void igs_setregs(int regtype, char* regs, int n);
void igs_getregs(int regtype, char* regs, int n);


#define	IGSVGA_MAJOR	100
static const char igsvga_name[] = "igsvga";
static volatile char* pVidPort;
static volatile unsigned short* pTvPort;


static struct file_operations igsvga_fops = {
	NULL,			/* lseek              */
	NULL,			/* read               */
	NULL,			/* write              */
	NULL,			/* no special readdir            */
	NULL,			/* no special select             */
	igsvga_ioctl,
	NULL,			/* no special mmap               */
	open_igsvga,
	release_igsvga,
	NULL,			/* no special fsync              */
	NULL,			/* no special fasync             */
	NULL,			/* no special check_media_change */
	NULL			/* no special revaldate          */
};



void	delay10ms(void)
{
	//wait 10 ms
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + 2;	//wait good 10 ms
	schedule();
}

void	delay1us(void)
{
	//wait 1 us
        inb(0x80);
}


void igs_getregs(int regtype, char* regs, int n)
{
	int i;
	unsigned short reg3c, reg33;
	volatile char c;

	for (i=0; i<n; i++)
	{
		switch(regtype)
		{
			case CRT_CMD:
				pVidPort[0x3d4] = i;
				udelay(10);
				regs[i] = pVidPort[0x3d5];
				break;
			case ATR_CMD:
				c = pVidPort[0x3da];
				udelay(10);
				pVidPort[0x3c0] = i;
				udelay(10);
				regs[i] = pVidPort[0x3c1];
				break;
			case SEQ_CMD:
				pVidPort[0x3c4] = i+1;
				udelay(10);
				regs[i] = pVidPort[0x3c5];
				break;
			case GRF_CMD:
			case EXT_CMD:
				pVidPort[0x3ce] = i;
				udelay(10);
				regs[i] = pVidPort[0x3cf];
				break;
			case MISC_CMD:
				regs[i] = pVidPort[0x3cc];
				break;

			case PAL_CMD:
				pVidPort[0x3ce] = 0x56;
				c = pVidPort[0x3cf];
				pVidPort[0x3cf] = c | 4;
				regs[i] = pVidPort[0x3c6];
				pVidPort[0x3cf] = c;
				break;

			case TV_CMD:
				pVidPort[0x3ce] = 0x33;
				reg33 = pVidPort[0x3cf];
				pVidPort[0x3cf] = reg33 | 0x9;
				pVidPort[0x3ce] = 0x3c;
				reg3c = pVidPort[0x3cf];
				pVidPort[0x3cf] = reg3c | 0x80;							
				regs[2*i] = *(char *)(&pTvPort[2*i]); /* Skip the odd addresses, do the read as two 8-bit reads as this is what the doc's say (16-bit write, though */
				regs[2*i + 1] = *((char *)(&pTvPort[2*i]) + 1);
/* Lock regs again		pVidPort[0x3ce] = 0x3c;
				pVidPort[0x3cf] = reg3c;
				pVidPort[0x3ce] = 0x33;
				pVidPort[0x3cf] = reg33;					
*/				
				break;
		}
	}
}

void igs_setregs(int regtype, char* regs, int n)
{
	int i;
	volatile char c;
	unsigned short reg3c, reg33;
	for (i=0; i<n; i++)
	{
		switch(regtype)
		{
			case CRT_CMD:
				pVidPort[0x3d4] = i;
				udelay(10);
				pVidPort[0x3d5] = regs[i];
				break;
			case ATR_CMD:
				c = pVidPort[0x3da];
				udelay(10);
				pVidPort[0x3c0] = i;
				udelay(10);
				pVidPort[0x3c0] = regs[i];
				break;
			case SEQ_CMD:
#if 0
				if (!i)
				{
					pVidPort[SEQ_I] = 0x01;
			    		pVidPort[SEQ_D] = regs[i] | 0x20;
				}
#endif
				pVidPort[0x3c4] = i+1;
				udelay(10);
				pVidPort[0x3c5] = regs[i];
				break;


			case GRF_CMD:
			case EXT_CMD:
				pVidPort[0x3ce] = i;
				udelay(10);
				pVidPort[0x3cf] = regs[i];
				break;
			case MISC_CMD:
				pVidPort[0x3c2] = regs[0];
				break;
			case TV_CMD:
				pVidPort[0x3ce] = 0x33;
				reg33 = pVidPort[0x3cf];
				pVidPort[0x3cf] = reg33 | 0x9;
				pVidPort[0x3ce] = 0x3c;
				reg3c = pVidPort[0x3cf];
				pVidPort[0x3cf] = reg3c | 0x80;							

				pTvPort[2*i] = ((unsigned short*)regs)[i];
				break;

			case PAL_CMD:
				printk("Setregs pal! %d\n", regs[0]);
				pVidPort[0x3ce] = 0xfa;
				pVidPort[0x3cf] = 0x05;
				pVidPort[0x3ce] = 0x56;
				c = pVidPort[0x3cf];
				pVidPort[0x3cf] = c | 4;
				pVidPort[0x3c6] = regs[0];
				pVidPort[0x3cf] = c;
				break;

		}
	}
}

void igs_setreg_specific(int regtype, int n, unsigned short w_val)
{
	char b_val = (char)(w_val & 0x00FF);
	volatile char c;
	unsigned short reg3c, reg33;
	switch(regtype)
	{
		case CRT_CMD:
			pVidPort[0x3d4] = n;
			udelay(10);
			pVidPort[0x3d5] = b_val;
			break;
		case ATR_CMD:
			c = pVidPort[0x3da];
			udelay(10);
			pVidPort[0x3c0] = n;
			udelay(10);
			pVidPort[0x3c0] = b_val;
			break;
		case SEQ_CMD:
#if 0
			if (!i)
			{
				pVidPort[SEQ_I] = 0x01;
			    	pVidPort[SEQ_D] = regs[i] | 0x20;
			}
#endif
			pVidPort[0x3c4] = n+1;
			udelay(10);
			pVidPort[0x3c5] = b_val;
			break;


		case GRF_CMD:
		case EXT_CMD:
			pVidPort[0x3ce] = n;
			udelay(10);
			pVidPort[0x3cf] = b_val;
			break;
		case MISC_CMD:
			pVidPort[0x3c2] = b_val;
			break;
		case TV_CMD:
			pVidPort[0x3ce] = 0x33;
			reg33 = pVidPort[0x3cf];
			pVidPort[0x3cf] = reg33 | 0x9;
			pVidPort[0x3ce] = 0x3c;
			reg3c = pVidPort[0x3cf];
			pVidPort[0x3cf] = reg3c | 0x80;							

//			printk("Setting reg %x = %x\n", 4*n+0xe430, w_val);
			pTvPort[2*n] = w_val;
			break;

		case PAL_CMD:
			printk("Setreg pal! %d\n", b_val);
			pVidPort[0x3ce] = 0x56;
			c = pVidPort[0x3cf];
			pVidPort[0x3cf] = c | 4;
			pVidPort[0x3c6] = b_val;
			pVidPort[0x3cf] = c;
			break;

	}
}

	
static int open_igsvga(struct inode* inodep, struct file* filep)
{
	MOD_INC_USE_COUNT;

	pVidPort = (volatile char*)0xd0800000;
	pTvPort = (volatile unsigned short*)(0xd08b0000 + 0xe430);

	return 0;
}


static void release_igsvga(struct inode* inodep, struct file* filep)
{
	MOD_DEC_USE_COUNT;
}


static int igsvga_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg)
{
long unsigned flags;
IGSREGS igs;
int	pArray;
volatile char c;
unsigned short reg33, reg3c;

//	printk("igsvga_ioctl: cmd = 0x%X, arg = 0x%X.\n",cmd, *(int*)arg);

        pArray = *(int *) arg;		//get the address of user IGSREGS

	switch (cmd)
	{
	case IGSVGA_WRITE_SPECIFIC:
	{
		struct reg_ops *r = (struct reg_ops *) arg;
		igs_setreg_specific(r->reg_t, r->reg, r->val);
		break;
	}
        case IGSVGA_WRITE:

            if (verify_area(VERIFY_READ, (void*)pArray, sizeof(IGSREGS) == -EFAULT))
		return(-EFAULT);

            memcpy_fromfs(&igs, (void*)pArray, sizeof(IGSREGS)); //get new regs

		save_flags(flags);
		cli();

//		c = pVidPort[0x3da];

	        /* synchronous reset on */
		pVidPort[SEQ_I] = 0x00;
		pVidPort[SEQ_D] = 0x01;

		igs_setregs(SEQ_CMD,(char*)&igs.seq_reg,4);
		igs_setregs(MISC_CMD,(char*)&igs.misc_reg,1);
		igs_setregs(PAL_CMD,(char*)&igs.pal_reg,1);

	        /* synchronous reset off */
	        pVidPort[SEQ_I] = 0x00;
		pVidPort[SEQ_D] = 0x03;


		/* Unprotect TV regs */
		pVidPort[0x3ce] = 0x33;
		reg33 = pVidPort[0x3cf];
		pVidPort[0x3cf] = reg33 | 0x9;
		pVidPort[0x3ce] = 0x3c;
		reg3c = pVidPort[0x3cf];
		pVidPort[0x3cf] = reg3c | 0x80;							
		igs_setregs(TV_CMD, (char*)&igs.tv_reg, 64);
		pVidPort[0x3ce] = 0x3c;
		pVidPort[0x3cf] = reg3c;
		pVidPort[0x3ce] = 0x33;
		pVidPort[0x3cf] = reg33;					
		

		/* Unprotect CRT regs 0..7 */
		pVidPort[CRT_I] = 0x11;
		pVidPort[CRT_D] = igs.crt_reg[0x11] & 0x7F;
		igs_setregs(CRT_CMD,(char*)&igs.crt_reg,25);

		igs_setregs(ATR_CMD,(char*)&igs.atr_reg,20);

		igs_setregs(GRF_CMD,(char*)&igs.grf_reg,256);

		pVidPort[0x3ce] = 0xB9;		//toggle this port to latch
		udelay(10);
		pVidPort[0x3cf] = 0x80;
		udelay(10);
		pVidPort[0x3cf] = 0x0;

		// turn video on
		c = pVidPort[0x3da];
		udelay(10);
		pVidPort[0x3c0] = 0x20;

		restore_flags(flags);

//	    printk("igsvga: Wrote IGS regs...\n");
	    break;

        case IGSVGA_READ:

		//first read all current values...
		save_flags(flags);
		cli();

		igs_getregs(GRF_CMD,(char*)&igs.grf_reg,256);
		igs_getregs(CRT_CMD,(char*)&igs.crt_reg,25);
//		c = pVidPort[0x3da];
		igs_getregs(ATR_CMD,(char*)&igs.atr_reg,20);
		igs_getregs(SEQ_CMD,(char*)&igs.seq_reg,4);
		igs_getregs(MISC_CMD,(char*)&igs.misc_reg,1);
		igs_getregs(TV_CMD, (char*)&igs.tv_reg, 64);
		igs_getregs(PAL_CMD, (char*)&igs.pal_reg, 1);

		// turn video on
		c = pVidPort[0x3da];
		udelay(10);
		pVidPort[0x3c0] = 0x20;

		restore_flags(flags);

		if (verify_area(VERIFY_WRITE, (void*)pArray, sizeof(IGSREGS) == -EFAULT))
			return(-EFAULT);

		memcpy_tofs((void*)pArray, &igs, sizeof(IGSREGS));
//		printk("igsvga: Returning IGS regs...\n");

		break;


	default:
		return -EINVAL;
	}

	return 0;
}


#ifndef MODULE
long igsvga_init(long mem_start, long mem_emd)
{
	if (igsvga_major = register_chrdev(0, igsvga_name, &igsvga_fops))
		printk("Unable to get major for igsvga device\n");
	else
		printk("igsvga V.0.0.2: major = %d... Woody 1998.\n", IGSVGA_MAJOR);

	return mem_start;
}
#else
int init_module(void)
{
int	temp;

//	printk("IGSVGA: init_module().\n");
//	nwprintk("IGSVGA: init_module().\n");
	if ((temp = register_chrdev(IGSVGA_MAJOR, igsvga_name, &igsvga_fops)) == -EBUSY)
	{
		printk("Unable to get major for igsvga device (%d)\n",temp);
		return -EIO;
	}
	else
	{
		printk("igsvga V.0.0.2: major = %d... Woody 1998.\n", IGSVGA_MAJOR);
//		nwprintk("igsvga V.0.0.2: major = %d... Woody 1998.\n", IGSVGA_MAJOR);
	}
	return 0;
}

void cleanup_module(void)
{
	unregister_chrdev(IGSVGA_MAJOR, igsvga_name);
}
#endif
